# -*- coding: utf-8 -*-
######################################################################################################################
# Copyright (C) 2017-2021 Spine project consortium
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

################################################################################
## Form generated from reading UI file 'data_store_properties.ui'
##
## Created by: Qt User Interface Compiler version 5.14.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide2.QtCore import (QCoreApplication, QDate, QDateTime, QMetaObject,
    QObject, QPoint, QRect, QSize, QTime, QUrl, Qt)
from PySide2.QtGui import (QBrush, QColor, QConicalGradient, QCursor, QFont,
    QFontDatabase, QIcon, QKeySequence, QLinearGradient, QPalette, QPainter,
    QPixmap, QRadialGradient)
from PySide2.QtWidgets import *

from spinetoolbox.widgets.custom_qlineedits import PropertyQLineEdit
from spinetoolbox.widgets.custom_qlineedits import CustomQLineEdit

from spine_items import resources_icons_rc

class Ui_Form(object):
    def setupUi(self, Form):
        if not Form.objectName():
            Form.setObjectName(u"Form")
        Form.resize(369, 337)
        self.verticalLayout = QVBoxLayout(Form)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.groupBox = QGroupBox(Form)
        self.groupBox.setObjectName(u"groupBox")
        self.gridLayout = QGridLayout(self.groupBox)
        self.gridLayout.setObjectName(u"gridLayout")
        self.gridLayout.setContentsMargins(3, 3, 3, 3)
        self.label_dialect = QLabel(self.groupBox)
        self.label_dialect.setObjectName(u"label_dialect")
        sizePolicy = QSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.label_dialect.sizePolicy().hasHeightForWidth())
        self.label_dialect.setSizePolicy(sizePolicy)
        self.label_dialect.setMaximumSize(QSize(16777215, 16777215))

        self.gridLayout.addWidget(self.label_dialect, 0, 0, 1, 1)

        self.lineEdit_port = PropertyQLineEdit(self.groupBox)
        self.lineEdit_port.setObjectName(u"lineEdit_port")
        self.lineEdit_port.setEnabled(False)
        sizePolicy1 = QSizePolicy(QSizePolicy.Expanding, QSizePolicy.Fixed)
        sizePolicy1.setHorizontalStretch(0)
        sizePolicy1.setVerticalStretch(0)
        sizePolicy1.setHeightForWidth(self.lineEdit_port.sizePolicy().hasHeightForWidth())
        self.lineEdit_port.setSizePolicy(sizePolicy1)
        self.lineEdit_port.setMinimumSize(QSize(0, 24))
        self.lineEdit_port.setMaximumSize(QSize(80, 24))
        self.lineEdit_port.setInputMethodHints(Qt.ImhNone)

        self.gridLayout.addWidget(self.lineEdit_port, 4, 5, 1, 1)

        self.label_dsn = QLabel(self.groupBox)
        self.label_dsn.setObjectName(u"label_dsn")
        sizePolicy.setHeightForWidth(self.label_dsn.sizePolicy().hasHeightForWidth())
        self.label_dsn.setSizePolicy(sizePolicy)

        self.gridLayout.addWidget(self.label_dsn, 1, 0, 1, 1)

        self.label_password = QLabel(self.groupBox)
        self.label_password.setObjectName(u"label_password")
        sizePolicy.setHeightForWidth(self.label_password.sizePolicy().hasHeightForWidth())
        self.label_password.setSizePolicy(sizePolicy)

        self.gridLayout.addWidget(self.label_password, 3, 0, 1, 1)

        self.label_host = QLabel(self.groupBox)
        self.label_host.setObjectName(u"label_host")
        sizePolicy.setHeightForWidth(self.label_host.sizePolicy().hasHeightForWidth())
        self.label_host.setSizePolicy(sizePolicy)

        self.gridLayout.addWidget(self.label_host, 4, 0, 1, 1)

        self.label_username = QLabel(self.groupBox)
        self.label_username.setObjectName(u"label_username")
        sizePolicy.setHeightForWidth(self.label_username.sizePolicy().hasHeightForWidth())
        self.label_username.setSizePolicy(sizePolicy)

        self.gridLayout.addWidget(self.label_username, 2, 0, 1, 1)

        self.label_port = QLabel(self.groupBox)
        self.label_port.setObjectName(u"label_port")
        sizePolicy2 = QSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed)
        sizePolicy2.setHorizontalStretch(1)
        sizePolicy2.setVerticalStretch(0)
        sizePolicy2.setHeightForWidth(self.label_port.sizePolicy().hasHeightForWidth())
        self.label_port.setSizePolicy(sizePolicy2)

        self.gridLayout.addWidget(self.label_port, 4, 4, 1, 1)

        self.label_database = QLabel(self.groupBox)
        self.label_database.setObjectName(u"label_database")
        sizePolicy.setHeightForWidth(self.label_database.sizePolicy().hasHeightForWidth())
        self.label_database.setSizePolicy(sizePolicy)

        self.gridLayout.addWidget(self.label_database, 6, 0, 1, 1)

        self.horizontalLayout_4 = QHBoxLayout()
        self.horizontalLayout_4.setObjectName(u"horizontalLayout_4")
        self.lineEdit_database = CustomQLineEdit(self.groupBox)
        self.lineEdit_database.setObjectName(u"lineEdit_database")
        self.lineEdit_database.setEnabled(False)
        sizePolicy1.setHeightForWidth(self.lineEdit_database.sizePolicy().hasHeightForWidth())
        self.lineEdit_database.setSizePolicy(sizePolicy1)
        self.lineEdit_database.setMinimumSize(QSize(0, 24))
        self.lineEdit_database.setMaximumSize(QSize(16777215, 24))
        self.lineEdit_database.setCursor(QCursor(Qt.IBeamCursor))
        self.lineEdit_database.setClearButtonEnabled(True)

        self.horizontalLayout_4.addWidget(self.lineEdit_database)

        self.toolButton_select_sqlite_file = QToolButton(self.groupBox)
        self.toolButton_select_sqlite_file.setObjectName(u"toolButton_select_sqlite_file")
        self.toolButton_select_sqlite_file.setEnabled(False)
        sizePolicy1.setHeightForWidth(self.toolButton_select_sqlite_file.sizePolicy().hasHeightForWidth())
        self.toolButton_select_sqlite_file.setSizePolicy(sizePolicy1)
        self.toolButton_select_sqlite_file.setMinimumSize(QSize(22, 22))
        self.toolButton_select_sqlite_file.setMaximumSize(QSize(22, 22))
        icon = QIcon()
        icon.addFile(u":/icons/folder-open-solid.svg", QSize(), QIcon.Normal, QIcon.Off)
        self.toolButton_select_sqlite_file.setIcon(icon)

        self.horizontalLayout_4.addWidget(self.toolButton_select_sqlite_file)


        self.gridLayout.addLayout(self.horizontalLayout_4, 6, 1, 1, 5)

        self.lineEdit_host = PropertyQLineEdit(self.groupBox)
        self.lineEdit_host.setObjectName(u"lineEdit_host")
        self.lineEdit_host.setEnabled(False)
        sizePolicy3 = QSizePolicy(QSizePolicy.Expanding, QSizePolicy.Fixed)
        sizePolicy3.setHorizontalStretch(3)
        sizePolicy3.setVerticalStretch(0)
        sizePolicy3.setHeightForWidth(self.lineEdit_host.sizePolicy().hasHeightForWidth())
        self.lineEdit_host.setSizePolicy(sizePolicy3)
        self.lineEdit_host.setMinimumSize(QSize(0, 24))
        self.lineEdit_host.setMaximumSize(QSize(5000, 24))
        self.lineEdit_host.setClearButtonEnabled(True)

        self.gridLayout.addWidget(self.lineEdit_host, 4, 1, 1, 3)

        self.lineEdit_password = PropertyQLineEdit(self.groupBox)
        self.lineEdit_password.setObjectName(u"lineEdit_password")
        self.lineEdit_password.setEnabled(False)
        sizePolicy1.setHeightForWidth(self.lineEdit_password.sizePolicy().hasHeightForWidth())
        self.lineEdit_password.setSizePolicy(sizePolicy1)
        self.lineEdit_password.setMinimumSize(QSize(0, 24))
        self.lineEdit_password.setMaximumSize(QSize(5000, 24))
        self.lineEdit_password.setEchoMode(QLineEdit.Password)
        self.lineEdit_password.setClearButtonEnabled(True)

        self.gridLayout.addWidget(self.lineEdit_password, 3, 1, 1, 5)

        self.lineEdit_username = PropertyQLineEdit(self.groupBox)
        self.lineEdit_username.setObjectName(u"lineEdit_username")
        self.lineEdit_username.setEnabled(False)
        sizePolicy1.setHeightForWidth(self.lineEdit_username.sizePolicy().hasHeightForWidth())
        self.lineEdit_username.setSizePolicy(sizePolicy1)
        self.lineEdit_username.setMinimumSize(QSize(0, 24))
        self.lineEdit_username.setMaximumSize(QSize(5000, 24))
        self.lineEdit_username.setClearButtonEnabled(True)

        self.gridLayout.addWidget(self.lineEdit_username, 2, 1, 1, 5)

        self.comboBox_dsn = QComboBox(self.groupBox)
        self.comboBox_dsn.setObjectName(u"comboBox_dsn")
        self.comboBox_dsn.setEnabled(False)
        sizePolicy1.setHeightForWidth(self.comboBox_dsn.sizePolicy().hasHeightForWidth())
        self.comboBox_dsn.setSizePolicy(sizePolicy1)
        self.comboBox_dsn.setMinimumSize(QSize(0, 24))
        self.comboBox_dsn.setMaximumSize(QSize(16777215, 24))

        self.gridLayout.addWidget(self.comboBox_dsn, 1, 1, 1, 5)

        self.comboBox_dialect = QComboBox(self.groupBox)
        self.comboBox_dialect.setObjectName(u"comboBox_dialect")
        sizePolicy1.setHeightForWidth(self.comboBox_dialect.sizePolicy().hasHeightForWidth())
        self.comboBox_dialect.setSizePolicy(sizePolicy1)
        self.comboBox_dialect.setMinimumSize(QSize(0, 24))
        self.comboBox_dialect.setMaximumSize(QSize(16777215, 24))

        self.gridLayout.addWidget(self.comboBox_dialect, 0, 1, 1, 5)


        self.verticalLayout.addWidget(self.groupBox)

        self.horizontalLayout_3 = QHBoxLayout()
        self.horizontalLayout_3.setObjectName(u"horizontalLayout_3")
        self.toolButton_vacuum = QPushButton(Form)
        self.toolButton_vacuum.setObjectName(u"toolButton_vacuum")
        icon1 = QIcon()
        icon1.addFile(u":/icons/broom.svg", QSize(), QIcon.Normal, QIcon.Off)
        self.toolButton_vacuum.setIcon(icon1)

        self.horizontalLayout_3.addWidget(self.toolButton_vacuum)

        self.horizontalSpacer_3 = QSpacerItem(40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum)

        self.horizontalLayout_3.addItem(self.horizontalSpacer_3)

        self.toolButton_copy_url = QPushButton(Form)
        self.toolButton_copy_url.setObjectName(u"toolButton_copy_url")
        icon2 = QIcon()
        icon2.addFile(u":/icons/copy.svg", QSize(), QIcon.Normal, QIcon.Off)
        self.toolButton_copy_url.setIcon(icon2)

        self.horizontalLayout_3.addWidget(self.toolButton_copy_url)


        self.verticalLayout.addLayout(self.horizontalLayout_3)

        self.verticalSpacer_2 = QSpacerItem(20, 40, QSizePolicy.Minimum, QSizePolicy.Expanding)

        self.verticalLayout.addItem(self.verticalSpacer_2)

        self.horizontalLayout = QHBoxLayout()
        self.horizontalLayout.setObjectName(u"horizontalLayout")
        self.pushButton_create_new_spine_db = QPushButton(Form)
        self.pushButton_create_new_spine_db.setObjectName(u"pushButton_create_new_spine_db")
        sizePolicy4 = QSizePolicy(QSizePolicy.MinimumExpanding, QSizePolicy.Fixed)
        sizePolicy4.setHorizontalStretch(0)
        sizePolicy4.setVerticalStretch(0)
        sizePolicy4.setHeightForWidth(self.pushButton_create_new_spine_db.sizePolicy().hasHeightForWidth())
        self.pushButton_create_new_spine_db.setSizePolicy(sizePolicy4)
        icon3 = QIcon()
        icon3.addFile(u":/icons/Spine_symbol.png", QSize(), QIcon.Normal, QIcon.Off)
        self.pushButton_create_new_spine_db.setIcon(icon3)

        self.horizontalLayout.addWidget(self.pushButton_create_new_spine_db)

        self.horizontalSpacer = QSpacerItem(40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum)

        self.horizontalLayout.addItem(self.horizontalSpacer)

        self.pushButton_ds_open_editor = QPushButton(Form)
        self.pushButton_ds_open_editor.setObjectName(u"pushButton_ds_open_editor")
        sizePolicy4.setHeightForWidth(self.pushButton_ds_open_editor.sizePolicy().hasHeightForWidth())
        self.pushButton_ds_open_editor.setSizePolicy(sizePolicy4)

        self.horizontalLayout.addWidget(self.pushButton_ds_open_editor)


        self.verticalLayout.addLayout(self.horizontalLayout)

        QWidget.setTabOrder(self.comboBox_dialect, self.comboBox_dsn)
        QWidget.setTabOrder(self.comboBox_dsn, self.lineEdit_username)
        QWidget.setTabOrder(self.lineEdit_username, self.lineEdit_password)
        QWidget.setTabOrder(self.lineEdit_password, self.lineEdit_host)
        QWidget.setTabOrder(self.lineEdit_host, self.lineEdit_port)
        QWidget.setTabOrder(self.lineEdit_port, self.lineEdit_database)
        QWidget.setTabOrder(self.lineEdit_database, self.toolButton_select_sqlite_file)
        QWidget.setTabOrder(self.toolButton_select_sqlite_file, self.toolButton_vacuum)
        QWidget.setTabOrder(self.toolButton_vacuum, self.toolButton_copy_url)
        QWidget.setTabOrder(self.toolButton_copy_url, self.pushButton_create_new_spine_db)
        QWidget.setTabOrder(self.pushButton_create_new_spine_db, self.pushButton_ds_open_editor)

        self.retranslateUi(Form)

        QMetaObject.connectSlotsByName(Form)
    # setupUi

    def retranslateUi(self, Form):
        Form.setWindowTitle(QCoreApplication.translate("Form", u"Form", None))
        self.groupBox.setTitle(QCoreApplication.translate("Form", u"URL", None))
        self.label_dialect.setText(QCoreApplication.translate("Form", u"Dialect", None))
        self.lineEdit_port.setPlaceholderText("")
        self.label_dsn.setText(QCoreApplication.translate("Form", u"DSN", None))
        self.label_password.setText(QCoreApplication.translate("Form", u"Password", None))
        self.label_host.setText(QCoreApplication.translate("Form", u"Host", None))
        self.label_username.setText(QCoreApplication.translate("Form", u"Username", None))
        self.label_port.setText(QCoreApplication.translate("Form", u"Port", None))
        self.label_database.setText(QCoreApplication.translate("Form", u"Database", None))
        self.lineEdit_database.setPlaceholderText("")
#if QT_CONFIG(tooltip)
        self.toolButton_select_sqlite_file.setToolTip(QCoreApplication.translate("Form", u"<html><head/><body><p>Select SQLite file</p></body></html>", None))
#endif // QT_CONFIG(tooltip)
        self.lineEdit_host.setPlaceholderText("")
        self.lineEdit_password.setPlaceholderText("")
        self.lineEdit_username.setPlaceholderText("")
        self.toolButton_vacuum.setText(QCoreApplication.translate("Form", u"Vacuum", None))
#if QT_CONFIG(tooltip)
        self.toolButton_copy_url.setToolTip(QCoreApplication.translate("Form", u"<html><head/><body><p>Copy current database url to clipboard.</p></body></html>", None))
#endif // QT_CONFIG(tooltip)
        self.toolButton_copy_url.setText(QCoreApplication.translate("Form", u"Copy URL", None))
#if QT_CONFIG(tooltip)
        self.pushButton_create_new_spine_db.setToolTip(QCoreApplication.translate("Form", u"<html><head/><body><p>Create new Spine database at the selected URL, or at a default one if the selected is not valid.</p></body></html>", None))
#endif // QT_CONFIG(tooltip)
        self.pushButton_create_new_spine_db.setText(QCoreApplication.translate("Form", u"New Spine db", None))
#if QT_CONFIG(tooltip)
        self.pushButton_ds_open_editor.setToolTip(QCoreApplication.translate("Form", u"<html><head/><body><p>Open URL in Spine database editor</p></body></html>", None))
#endif // QT_CONFIG(tooltip)
        self.pushButton_ds_open_editor.setText(QCoreApplication.translate("Form", u"Open editor...", None))
    # retranslateUi




######################################################################################################################
# Copyright (C) 2017-2021 Spine project consortium
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""
Parameter indexing settings window for .gdx export.

:author: A. Soininen (VTT)
:date:   25.11.2019
"""
from contextlib import contextmanager
from copy import deepcopy
from PySide2.QtCore import QItemSelectionModel, QModelIndex, Qt, Signal, Slot
from PySide2.QtGui import QStandardItem
from PySide2.QtWidgets import QMessageBox, QWidget
from spinedb_api.spine_io.exporters import gdx
from spinedb_api import DatabaseMapping
from .parameter_index_settings import IndexSettingsState, ParameterIndexSettings
from ..mvcmodels.indexing_domain_list_model import IndexingDomainListModel

_PARAMETER_ROLE = Qt.UserRole + 1
_PARAMETER_NAME_ROLE = Qt.UserRole + 2


class ParameterIndexSettingsWindow(QWidget):
    """A window which shows a list of ParameterIndexSettings widgets, one for each parameter with indexed values."""

    settings_approved = Signal()
    """Emitted when the settings have been approved."""
    settings_rejected = Signal()
    """Emitted when the settings have been rejected."""

    def __init__(self, indexing_settings, set_settings, database_path, none_fallback, parent):
        """
        Args:
            indexing_settings (dict): a map from parameter name to a dict of domain names and :class:`IndexingSetting`
            set_settings (SetSettings): export settings
            database_path (str): a database url
            none_fallback (NoneFallback): how to handle None values
            parent (QWidget): a parent widget
        """
        from ..ui.parameter_index_settings_window import Ui_Form  # pylint: disable=import-outside-toplevel

        super().__init__(parent, f=Qt.Window)
        self._set_settings = set_settings
        self._database_mapping = DatabaseMapping(database_path) if database_path else None
        self._enable_domain_updates = True
        self._parameters = gdx.indexed_parameters(self._database_mapping, none_fallback, logger=None)
        self._indexing_settings = self._fix_legacy_indexing_settings(indexing_settings)
        self._ui = Ui_Form()
        self._ui.setupUi(self)
        self.setWindowTitle(f"Gdx Parameter Indexing Settings")
        self.setAttribute(Qt.WA_DeleteOnClose, True)
        self._ui.splitter.setSizes([400, 50])
        self._ui.button_box.accepted.connect(self._collect_and_hide)
        self._ui.button_box.rejected.connect(self._reject_and_close)
        self._additional_domains_model = IndexingDomainListModel(set_settings, self._parameters)
        self._additional_domains_model.domain_renamed.connect(self._update_after_domain_rename)
        self._additional_domains_model.domains_added.connect(self._add_available_domains)
        self._additional_domains_model.domains_removed.connect(self._remove_available_domains)
        self._ui.additional_domains_list_view.setModel(self._additional_domains_model)
        self._ui.additional_domains_list_view.selectionModel().currentChanged.connect(self._load_additional_domain)
        self._ui.add_domain_button.clicked.connect(self._add_domain)
        self._ui.remove_domain_button.clicked.connect(self._remove_selected_domains)
        self._ui.description_edit.textEdited.connect(self._set_domain_description)
        self._ui.use_expression_radio_button.clicked.connect(self._use_expression)
        self._ui.expression_edit.textChanged.connect(self._update_expression)
        self._ui.length_spin_box.valueChanged.connect(self._update_length)
        self._ui.extract_from_radio_button.clicked.connect(self._use_extraction)
        self._set_additional_domain_widgets_enabled(False)
        self._populate_extract_from_combo()
        self._ui.extract_from_combo_box.currentIndexChanged.connect(self._set_extraction_domain)
        self._settings_widgets = dict()
        self._available_domains = {name: set_settings.records(name) for name in set_settings.domain_names}
        for parameter_name, by_dimensions in self._indexing_settings.items():
            parameter_by_domains = self._parameters.get(parameter_name, {})
            for domain_names, indexing_setting in by_dimensions.items():
                parameter = parameter_by_domains.get(domain_names)
                if parameter is None:
                    parameter = gdx.Parameter(domain_names, [], [])
                settings_widget = ParameterIndexSettings(
                    parameter_name,
                    parameter,
                    indexing_setting,
                    self._available_domains,
                    self._ui.settings_area_contents,
                )
                self._ui.settings_area_layout.insertWidget(0, settings_widget)
                widgets = self._settings_widgets.setdefault(parameter_name, dict())
                widgets[domain_names] = settings_widget
        if not self._indexing_settings:
            self._ui.widget_stack.setCurrentIndex(1)
            return
        self._ui.widget_stack.setCurrentIndex(0)

    @property
    def indexing_settings(self):
        """indexing settings dictionary"""
        return self._indexing_settings

    def additional_indexing_domains(self):
        return self._additional_domains_model.gather_domains(self._parameters)

    def set_domain_updated_enabled(self, enabled):
        """
        Enables or disables updating the indexing settings widgets.

        Args:
            enabled (bool): if True, allow the widgets to update
        """
        self._enable_domain_updates = enabled

    def _switch_additional_domain_widgets_enabled_state(self, using_expression):
        """
        Enabled and disables additional domain widgets.

        Args:
            using_expression (bool): True if expression is used,
                False if record keys are extracted from existing parameter
        """
        self._ui.expression_edit.setEnabled(using_expression)
        self._ui.length_spin_box.setEnabled(using_expression)
        self._ui.extract_from_combo_box.setEnabled(not using_expression)

    def _set_additional_domain_widgets_enabled(self, enabled):
        self._ui.description_edit.setEnabled(enabled)
        self._ui.use_expression_radio_button.setEnabled(enabled)
        self._ui.expression_edit.setEnabled(enabled)
        self._ui.extract_from_radio_button.setEnabled(enabled)
        self._ui.extract_from_combo_box.setEnabled(enabled)

    @Slot(bool)
    def _add_domain(self, _):
        """Creates a new additional domain."""
        self._additional_domains_model.create_new_domain()
        new_current = self._additional_domains_model.index(self._additional_domains_model.rowCount() - 1, 0)
        self._ui.additional_domains_list_view.selectionModel().setCurrentIndex(
            new_current, QItemSelectionModel.ClearAndSelect
        )

    @Slot()
    def _collect_and_hide(self):
        """Collects settings from individual ParameterIndexSettings widgets and hides the window."""
        for parameter_name, widgets in self._settings_widgets.items():
            for domain_names, settings_widget in widgets.items():
                if settings_widget.state != IndexSettingsState.OK:
                    self._ui.settings_area.ensureWidgetVisible(settings_widget)
                    message = f"Parameter '{parameter_name}' indexing not well-defined."
                    QMessageBox.warning(self, "Bad Parameter Indexing", message)
                    return
        for parameter_name, widgets in self._settings_widgets.items():
            for domain_names, settings_widget in widgets.items():
                setting = self._indexing_settings[parameter_name][domain_names]
                setting.indexing_domain_name = settings_widget.indexing_domain_name()
                setting.picking = settings_widget.picking()
        self.settings_approved.emit()
        self.hide()

    @Slot(QModelIndex, QModelIndex)
    def _load_additional_domain(self, current, previous):
        if not previous.isValid():
            self._set_additional_domain_widgets_enabled(True)
        if not current.isValid():
            self._set_additional_domain_widgets_enabled(False)
            return
        with _disable_domain_updates(self):
            domain_proto = self._additional_domains_model.item_at(current.row())
            self._ui.description_edit.setText(domain_proto.description)
            if domain_proto.expression is not None:
                self._ui.use_expression_radio_button.setChecked(True)
                self._ui.expression_edit.setText(domain_proto.expression)
                self._ui.length_spin_box.setValue(domain_proto.length)
                self._ui.extract_from_combo_box.setCurrentIndex(-1)
                self._switch_additional_domain_widgets_enabled_state(True)
            else:
                self._ui.extract_from_radio_button.setChecked(True)
                self._ui.extract_from_combo_box.setCurrentIndex(-1)
                if domain_proto.extract_from_parameter_name is not None:
                    model = self._ui.extract_from_combo_box.model()
                    for row in range(model.rowCount()):
                        index = model.index(row, 0)
                        parameter_name = index.data(_PARAMETER_NAME_ROLE)
                        if domain_proto.extract_from_parameter_name != parameter_name:
                            continue
                        parameter = index.data(_PARAMETER_ROLE)
                        if domain_proto.extract_from_parameter_domain_names == parameter.domain_names:
                            self._ui.extract_from_combo_box.setCurrentIndex(row)
                            break
                self._ui.expression_edit.clear()
                self._switch_additional_domain_widgets_enabled_state(False)

    @Slot()
    def _reject_and_close(self):
        self.close()

    @Slot(bool)
    def _remove_selected_domains(self, _):
        selection_model = self._ui.additional_domains_list_view.selectionModel()
        if not selection_model.hasSelection():
            return
        rows = [index.row() for index in selection_model.selectedRows()]
        self._additional_domains_model.remove_rows(rows)
        current = selection_model.currentIndex()
        if current.isValid():
            selection_model.select(current, QItemSelectionModel.ClearAndSelect)

    @Slot(list)
    def _remove_available_domains(self, removed_domains):
        """
        Updates available domains.

        Args:
            removed_domains (list of str): domains that were removed
        """
        map(self._available_domains.pop, removed_domains)
        for widgets in self._settings_widgets.values():
            for widget in widgets.values():
                for domain in removed_domains:
                    widget.remove_domain(domain)

    @Slot(list)
    def _add_available_domains(self, new_domains):
        """
        Updates available domains.

        Args:
            new_domains (list of str): a list of new domain names
        """

        self._available_domains.update(self._additional_domains_model.gather_domains(self._parameters)[0])
        for widgets in self._settings_widgets.values():
            for widget in widgets.values():
                for domain in new_domains:
                    widget.add_domain(domain)

    @Slot(str, str)
    def _update_after_domain_rename(self, old_name, new_name):
        """
        Propagates changes in domain names to widgets.

        Args:
            old_name (str): domain's previous name
            new_name (str): domain's current name
        """
        self._available_domains[new_name] = self._available_domains.pop(old_name)
        for widgets in self._settings_widgets.values():
            for widget in widgets.values():
                widget.update_domain_name(old_name, new_name)

    @Slot(str)
    def _update_expression(self, expression):
        """
        Updates the domain's record key expression.

        Args:
            expression (str): new expression
        """
        list_index = self._ui.additional_domains_list_view.currentIndex()
        if not list_index.isValid() or not self._enable_domain_updates:
            return
        item = self._additional_domains_model.item_at(list_index.row())
        item.expression = expression
        records = gdx.GeneratedRecords(item.expression, item.length)
        self._available_domains[item.name] = records
        for widgets in self._settings_widgets.values():
            for widget in widgets.values():
                widget.update_records(item.name)

    @Slot(int)
    def _update_length(self, length):
        """
        Updates the number of additional domain's records.

        Args:
            length (int): new record count
        """
        list_index = self._ui.additional_domains_list_view.currentIndex()
        if not list_index.isValid() or not self._enable_domain_updates:
            return
        item = self._additional_domains_model.item_at(list_index.row())
        item.length = length
        records = gdx.GeneratedRecords(item.expression, item.length)
        self._available_domains[item.name] = records
        for widgets in self._settings_widgets.values():
            for widget in widgets.values():
                widget.update_records(item.name)

    @Slot(bool)
    def _use_expression(self, _):
        self._switch_additional_domain_widgets_enabled_state(True)
        list_index = self._ui.additional_domains_list_view.currentIndex()
        if not list_index.isValid():
            return
        item = self._additional_domains_model.item_at(list_index.row())
        item.expression = self._ui.expression_edit.text()
        item.length = self._ui.length_spin_box.value()
        item.extract_from_parameter_name = None
        item.extract_from_parameter_domain_names = None
        records = gdx.GeneratedRecords(item.expression, item.length)
        self._available_domains[item.name] = records
        for widgets in self._settings_widgets.values():
            for widget in widgets.values():
                widget.update_records(item.name)

    @Slot(bool)
    def _use_extraction(self, _):
        self._switch_additional_domain_widgets_enabled_state(False)
        index = self._ui.extract_from_combo_box.currentIndex()
        self._set_extraction_domain(index)

    @Slot(str)
    def _set_extraction_domain(self, index):
        """
        Sets the domain from which domain's records are extracted.

        Args:
            index (int): choice in the extract from combo box
        """
        if not self._enable_domain_updates or index < 0:
            return
        domain_list_index = self._ui.additional_domains_list_view.currentIndex()
        if not domain_list_index.isValid():
            return
        model = self._ui.extract_from_combo_box.model()
        model_index = model.index(index, 0)
        parameter = model_index.data(_PARAMETER_ROLE)
        parameter_name = model_index.data(_PARAMETER_NAME_ROLE)
        item = self._additional_domains_model.item_at(domain_list_index.row())
        item.expression = None
        item.extract_from_parameter_name = parameter_name
        item.extract_from_parameter_domain_names = parameter.domain_names
        first_value = next(iter(parameter.values))
        value_indexes = [(str(i),) for i in first_value.indexes]
        records = gdx.ExtractedRecords(parameter_name, parameter.domain_names, value_indexes)
        self._available_domains[item.name] = records
        for widgets in self._settings_widgets.values():
            for widget in widgets.values():
                widget.update_records(item.name)

    @Slot(str)
    def _set_domain_description(self, description):
        """
        Sets currently selected additional domain's description.

        Args:
            description (str): description text
        """
        domain_list_index = self._ui.additional_domains_list_view.currentIndex()
        item = self._additional_domains_model.item_at(domain_list_index.row())
        item.description = description

    def closeEvent(self, event):
        """Handles the close event."""
        super().closeEvent(event)
        if self._database_mapping is not None:
            self._database_mapping.connection.close()
        self.settings_rejected.emit()

    def _populate_extract_from_combo(self):
        """Populates the extract from combo box."""
        model = self._ui.extract_from_combo_box.model()
        items = list()
        for parameter_name, by_dimensions in self._parameters.items():
            if not by_dimensions:
                continue
            item = QStandardItem()
            item.setData(next(iter(by_dimensions.values())), _PARAMETER_ROLE)
            item.setData(parameter_name, _PARAMETER_NAME_ROLE)
            if len(by_dimensions) == 1:
                item.setText(parameter_name)
            else:
                for domain_names in by_dimensions:
                    if len(domain_names) == 1:
                        item.setText(f"{parameter_name} ({domain_names[0]})")
                    else:
                        item.setText(f"{parameter_name} {domain_names}")
            items.append(item)
        for item in sorted(items, key=lambda i: i.text()):
            model.appendRow(item)

    def _fix_legacy_indexing_settings(self, indexing_settings):
        """
        Fixes indexing settings that are loaded from 0.5 project file and are missing parameters' domain names.

        Args:
            indexing_settings (dict): indexing settings

        Returns:
            dict: fixed indexing settings
        """
        for parameter_name, by_dimensions in indexing_settings.items():
            if next(iter(by_dimensions))[0] is None:
                setting = next(iter(by_dimensions.values()))
                by_dimensions.clear()
                for domain_names in self._parameters[parameter_name]:
                    by_dimensions[domain_names] = deepcopy(setting)
        return indexing_settings


@contextmanager
def _disable_domain_updates(window):
    """
    A context manager which disables updates on the indexing settings widgets.

    Args:
        window (ParameterIndexSettingsWindow): settings window
    """
    window.set_domain_updated_enabled(False)
    try:
        yield None
    finally:
        window.set_domain_updated_enabled(True)



######################################################################################################################
# Copyright (C) 2017-2021 Spine project consortium
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""
Classes for custom context menus and pop-up menus for Gimlets

:author: P. Savolainen (VTT)
:date:   15.4.2020
"""

from spinetoolbox.widgets.custom_menus import CustomContextMenu


class GimletPropertiesContextMenu(CustomContextMenu):
    """Context menu class for a Gimlet project item properties."""

    def __init__(self, parent, position, index):
        """

        Args:
            parent (QWidget): Parent for menu widget (ToolboxUI)
            position (QPoint): Position on screen
            index (QModelIndex): Index of item that requested the context-menu
        """
        super().__init__(parent, position)
        if not index.isValid():
            # If no item at index
            return
        self.add_action("Open Gimlet")



######################################################################################################################
# Copyright (C) 2017-2021 Spine project consortium
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""
Importer's execute kernel (do_work), as target for a multiprocess.Process

:authors: M. Marin (KTH)
:date:   6.11.2020
"""

import os
import spinedb_api
from spinedb_api.import_mapping.type_conversion import value_to_convert_spec
from spine_engine.utils.helpers import create_log_file_timestamp


def do_work(
    mapping, cancel_on_error, purge_before_writing, on_conflict, logs_dir, sources, connector, urls_downstream, logger
):
    all_data = []
    all_errors = []
    table_mappings = {
        name: mappings
        for name, mappings in mapping.get("table_mappings", {}).items()
        if name in mapping["selected_tables"]
    }
    table_options = {
        name: options
        for name, options in mapping.get("table_options", {}).items()
        if name in mapping["selected_tables"]
    }
    table_column_convert_specs = {
        tn: {int(col): value_to_convert_spec(spec) for col, spec in cols.items()}
        for tn, cols in mapping.get("table_types", {}).items()
    }
    table_row_convert_specs = {
        tn: {int(col): value_to_convert_spec(spec) for col, spec in cols.items()}
        for tn, cols in mapping.get("table_row_types", {}).items()
    }
    for src in sources:
        logger.msg.emit("Importing " + src)
        try:
            connector.connect_to_source(src)
        except Exception as error:  # pylint: disable=broad-except
            logger.msg_error.emit(f"Failed to connect to {src}: {error}")
            return (False,)
        for name, mappings in table_mappings.items():
            logger.msg.emit(f"Processing table <b>{name}</b>")
            for spec in mappings:
                mapping_name = next(iter(spec.keys()))
                logger.msg.emit(f"* Applying mapping <b>{mapping_name}</b>...")
                try:
                    data, errors = connector.get_mapped_data(
                        {name: [spec]}, table_options, table_column_convert_specs, table_row_convert_specs
                    )
                except spinedb_api.InvalidMapping as error:
                    logger.msg_error.emit(f"Failed to import: {error}")
                    if cancel_on_error:
                        logger.msg_error.emit("Cancel import on error has been set. Bailing out.")
                        return (False,)
                    logger.msg_warning.emit("Ignoring errors. Set Cancel import on error to bail out instead.")
                    continue
                if not errors:
                    logger.msg.emit(f"Successful ({sum(len(d) for d in data.values())} data to be written).")
                else:
                    logger.msg_warning.emit(
                        f"Read {sum(len(d) for d in data.values())} data with {len(errors)} errors."
                    )
                all_data.append(data)
                all_errors.extend(errors)
    if all_errors:
        # Log errors in a time stamped file into the logs directory
        timestamp = create_log_file_timestamp()
        logfilepath = os.path.abspath(os.path.join(logs_dir, timestamp + "_read_error.log"))
        with open(logfilepath, "w") as f:
            for err in all_errors:
                f.write(f"{err}\n")
        # Make error log file anchor with path as tooltip
        logfile_anchor = (
            "<a style='color:#BB99FF;' title='" + logfilepath + "' href='file:///" + logfilepath + "'>Error log</a>"
        )
        logger.msg_error.emit(logfile_anchor)
        if cancel_on_error:
            logger.msg_error.emit("Cancel import on error has been set. Bailing out.")
            return (False,)
        logger.msg_warning.emit("Ignoring errors. Set Cancel import on error to bail out instead.")
    if all_data:
        for url in urls_downstream:
            success = _import_data_to_url(
                cancel_on_error, purge_before_writing, on_conflict, logs_dir, all_data, url, logger
            )
            if not success and cancel_on_error:
                return (False,)
    return (True,)


def _import_data_to_url(cancel_on_error, purge_before_writing, on_conflict, logs_dir, all_data, url, logger):
    if purge_before_writing:
        spinedb_api.create_new_spine_database(url)
    try:
        db_map = spinedb_api.DatabaseMapping(url, upgrade=False, username="Importer")
    except (spinedb_api.SpineDBAPIError, spinedb_api.SpineDBVersionError) as err:
        logger.msg_error.emit(f"Unable to create database mapping, all import operations will be omitted: {err}")
        return False
    all_import_errors = []
    all_import_num = 0
    for data in all_data:
        import_num, import_errors = spinedb_api.import_data(db_map, on_conflict=on_conflict, **data)
        all_import_errors += import_errors
        if import_errors:
            logger.msg_error.emit("Errors while importing a table.")
            if cancel_on_error:
                logger.msg_error.emit("Cancel import on error is set. Bailing out.")
                if db_map.has_pending_changes():
                    logger.msg_error.emit("Rolling back changes.")
                    db_map.rollback_session()
                break
            logger.msg_warning.emit("Ignoring errors. Set Cancel import on error to bail out instead.")
        all_import_num += import_num
    if db_map.has_pending_changes():
        db_map.commit_session("Import data by Spine Toolbox Importer")
        logger.msg_success.emit(
            f"Inserted {all_import_num} data with {len(all_import_errors)} errors into {db_map.codename}"
        )
    else:
        logger.msg_warning.emit("No new data imported")
    db_map.connection.close()
    if all_import_errors:
        # Log errors in a time stamped file into the logs directory
        timestamp = create_log_file_timestamp()
        logfilepath = os.path.abspath(os.path.join(logs_dir, timestamp + "_import_error.log"))
        with open(logfilepath, "w") as f:
            for err in all_import_errors:
                f.write(str(err) + "\n")
        # Make error log file anchor with path as tooltip
        logfile_anchor = (
            "<a style='color:#BB99FF;' title='" + logfilepath + "' href='file:///" + logfilepath + "'>Error log</a>"
        )
        logger.msg_error.emit(logfile_anchor)
        return False
    return True
